@@ -1,23 +1,22 @@
 import json as _json
+import math as _math
+import multiprocessing as _mp
 import os as _os
-import numpy as _np
+from functools import partial as _partial
+from itertools import chain as _chain
+from os import sep as _sep
+from typing import List as _List, Union as _Union
 
+import numpy as _np
 import torch as _torch
+import torch.distributed as _dist
+import torch.utils.data as _data
 from torch.utils.data import DataLoader as _DataLoader, Dataset as _Dataset
 from torch.utils.data._utils.collate import default_collate as _default_collate
+
+import easytorch.data.datautils as _du
+import easytorch.utils as _etutils
 from easytorch.utils.logger import *
-from os import sep as _sep
-from typing import List as _List
-import torch.utils.data as _data
-import torch.distributed as _dist
-import math as _math
-
-
-def safe_collate(batch):
-    r"""
-    Savely select batches/skip errors in file loading.
-    """
-    return _default_collate([b for b in batch if b])
 
 
 def seed_worker(worker_id):
@@ -25,23 +24,118 @@ def seed_worker(worker_id):
     _np.random.seed(seed)
 
 
+def load_sparse_data_worker(dataset_cls, mode, args, dataspec, total, i, file):
+    test_dataset = dataset_cls(mode=mode, **args)
+    test_dataset.add(files=[file], verbose=False, **dataspec)
+    if args['verbose']:
+        print(f"Data items loaded: [ {i} / {total} ]", end='\r')
+    return [test_dataset]
+
+
+def safe_collate(batch):
+    r"""
+    Savely select batches/skip dataset_cls(errors in file loading.
+    """
+    return _default_collate([b for b in batch if b])
+
+
+def num_workers(args, loader_args, distributed=False):
+    if distributed:
+        return (loader_args['num_workers'] + args['num_gpus'] - 1) // args['num_gpus']
+    return loader_args['num_workers']
+
+
+def batch_size(args, loader_args, distributed=False):
+    if distributed:
+        loader_args['batch_size'] = loader_args['batch_size'] // args['num_gpus']
+    return loader_args['batch_size']
+
+
 class ETDataHandle:
 
     def __init__(self, args=None, dataloader_args=None, **kw):
-        self.args = {**args}
         self.dataset = {}
         self.dataloader = {}
-        self.dataloader_args = {}
-        if dataloader_args is not None:
-            self.dataloader_args.update(**dataloader_args)
-        self.args.update(**kw)
+        self.args = _etutils.FrozenDict(args)
+        self.dataloader_args = _etutils.FrozenDict(dataloader_args)
+
+    def get_dataset(self, handle_key, files, dataspec: dict, dataset_cls=None) -> _Dataset:
+        dataset = dataset_cls(mode=handle_key, limit=self.args['load_limit'], **self.args)
+        dataset.add(files=files, verbose=self.args['verbose'], **dataspec)
+        self.dataset[handle_key] = dataset
+        return dataset
+
+    def get_sparse_dataset(self, handle_key, files, dataspec: dict, dataset_cls=None) -> _List[_Dataset]:
+        r"""
+        Load the test data from current fold/split.
+        If -sp/--load-sparse arg is set, we need to load one image in one dataloader.
+        So that we can correctly gather components of one image(components like output patches)
+        """
+        _files = []
+        for i, file in enumerate(files[:self.args['load_limit']], 1):
+            _files.append([i, file])
+
+        sparse_datasets = []
+        if len(_files) > 0:
+            nw = min(num_workers(self.args, self.args, self.args['use_ddp']), len(_files))
+            with _mp.Pool(processes=max(1, nw)) as pool:
+                sparse_datasets = list(
+                    _chain.from_iterable(pool.starmap(
+                        _partial(
+                            load_sparse_data_worker, dataset_cls, handle_key, self.args, dataspec, len(_files)
+                        ),
+                        _files
+                    )))
+
+        success(f"\n{dataspec['name']}, {handle_key}, {len(sparse_datasets)} sparse dataset Loaded",
+                self.args['verbose'])
+        return sparse_datasets
+
+    def get_train_dataset(self, split_file, dataspec: dict, dataset_cls=None) -> _Union[_Dataset, _List[_Dataset]]:
+        if dataset_cls is None or self.dataloader_args.get('train', {}).get('dataset'):
+            return self.dataloader_args.get('train', {}).get('dataset')
+
+        r"""Load the train data from current fold/split."""
+        with open(dataspec['split_dir'] + _sep + split_file) as file:
+            split = _json.loads(file.read())
+            train_dataset = self.get_dataset('train', split.get('train', []),
+                                             dataspec, dataset_cls=dataset_cls)
+            return train_dataset
+
+    def get_validation_dataset(self, split_file, dataspec: dict, dataset_cls=None) -> _Union[_Dataset, _List[_Dataset]]:
+        if dataset_cls is None or self.dataloader_args.get('validation', {}).get('dataset'):
+            return self.dataloader_args.get('validation', {}).get('dataset')
+
+        r""" Load the validation data from current fold/split."""
+        with open(dataspec['split_dir'] + _sep + split_file) as file:
+            split = _json.loads(file.read())
+            val_dataset = self.get_dataset('validation', split.get('validation', []),
+                                           dataspec, dataset_cls=dataset_cls)
+            if val_dataset and len(val_dataset) > 0:
+                return val_dataset
+
+    def get_test_dataset(self, split_file, dataspec: dict, dataset_cls=None) -> _Union[_Dataset, _List[_Dataset]]:
+        if dataset_cls is None or self.dataloader_args.get('test', {}).get('dataset'):
+            return self.dataloader_args.get('test', {}).get('dataset')
+
+        with open(dataspec['split_dir'] + _sep + split_file) as file:
+            split = _json.loads(file.read())
+            _files = split.get('test', [])
+
+            if len(_files) > 0 and self.args.get('load_sparse'):
+                datasets = self.get_sparse_dataset('test', _files, dataspec, dataset_cls)
+            else:
+                datasets = self.get_dataset('test', _files, dataspec, dataset_cls=dataset_cls)
+
+            if len(datasets) > 0 and sum([len(t) for t in datasets if t]) > 0:
+                return datasets
 
     def get_loader(self, handle_key='', distributed=False, use_unpadded_sampler=False, **kw) -> _DataLoader:
         args = {**self.args}
         args['distributed'] = distributed
         args['use_unpadded_sampler'] = use_unpadded_sampler
-        args.update(**kw)
         args.update(self.dataloader_args.get(handle_key, {}))
+        args.update(**kw)
 
         loader_args = {
             'dataset': None,
@@ -74,74 +168,41 @@ class ETDataHandle:
                     loader_args['sampler'] = _data.distributed.DistributedSampler(loader_args['dataset'],
                                                                                   **sampler_args)
 
-            loader_args['num_workers'] = (loader_args['num_workers'] + args['num_gpus'] - 1) // args['num_gpus']
-            loader_args['batch_size'] = loader_args['batch_size'] // args['num_gpus']
+            loader_args['num_workers'] = num_workers(args, loader_args, True)
+            loader_args['batch_size'] = batch_size(args, loader_args, True)
 
         self.dataloader[handle_key] = _DataLoader(collate_fn=safe_collate, **loader_args)
         return self.dataloader[handle_key]
 
-    def get_dataset(self, handle_key, files, dataspec: dict, dataset_cls=None) -> _Dataset:
-        dataset = dataset_cls(mode=handle_key, limit=self.args['load_limit'], **self.args)
-        dataset.add(files=files, verbose=self.args['verbose'], **dataspec)
-        self.dataset[handle_key] = dataset
-        return dataset
+    def create_splits(self, dataspec, out_dir):
+        if _du.should_create_splits_(out_dir, dataspec, self.args):
+            _du.default_data_splitter_(dspec=dataspec, args=self.args)
+            info(f"{len(_os.listdir(dataspec['split_dir']))} split(s) created in '{dataspec['split_dir']}' directory.",
+                 self.args['verbose'])
+        else:
+            splits_len = len(_os.listdir(dataspec['split_dir']))
+            info(f"{splits_len} split(s) loaded from '{dataspec['split_dir']}' directory.",
+                 self.args['verbose'] and splits_len > 0)
 
-    def get_train_dataset(self, split_file, dataspec: dict, dataset_cls=None) -> _Dataset:
-        if dataset_cls is None or self.dataloader_args.get('train', {}).get('dataset'):
-            return self.dataloader_args.get('train', {}).get('dataset')
-
-        r"""Load the train data from current fold/split."""
-        with open(dataspec['split_dir'] + _sep + split_file) as file:
-            split = _json.loads(file.read())
-            train_dataset = self.get_dataset('train', split.get('train', []),
-                                             dataspec, dataset_cls=dataset_cls)
-            return train_dataset
-
-    def get_validation_dataset(self, split_file, dataspec: dict, dataset_cls=None) -> _Dataset:
-        if dataset_cls is None or self.dataloader_args.get('validation', {}).get('dataset'):
-            return self.dataloader_args.get('validation', {}).get('dataset')
-
-        r""" Load the validation data from current fold/split."""
-        with open(dataspec['split_dir'] + _sep + split_file) as file:
-            split = _json.loads(file.read())
-            val_dataset = self.get_dataset('validation', split.get('validation', []),
-                                           dataspec, dataset_cls=dataset_cls)
-            if val_dataset and len(val_dataset) > 0:
-                return val_dataset
-
-    def get_test_dataset(self, split_file, dataspec: dict, dataset_cls=None) -> _List[_Dataset]:
-        if dataset_cls is None or self.dataloader_args.get('test', {}).get('dataset'):
-            return self.dataloader_args.get('test', {}).get('dataset')
-
-        r"""
-        Load the test data from current fold/split.
-        If -sp/--load-sparse arg is set, we need to load one image in one dataloader.
-        So that we can correctly gather components of one image(components like output patches)
-        """
-        test_dataset_list = []
-        with open(dataspec['split_dir'] + _sep + split_file) as file:
-            files = _json.loads(file.read()).get('test', [])
-            if self.args.get('load_sparse'):
-                for f in files:
-                    if self.args['load_limit'] and len(test_dataset_list) >= self.args['load_limit']:
-                        break
-                    test_dataset = dataset_cls(mode='test', limit=self.args['load_limit'], **self.args)
-                    test_dataset.add(files=[f], verbose=False, **dataspec)
-                    test_dataset_list.append(test_dataset)
-                success(f'{len(test_dataset_list)} sparse dataset loaded.', self.args['verbose'])
-            else:
-                test_dataset_list.append(self.get_dataset('test', files, dataspec, dataset_cls=dataset_cls))
-
-        if len(test_dataset_list) > 0 and sum([len(t) for t in test_dataset_list if t]) > 0:
-            return test_dataset_list
+    def init_dataspec_(self, dataspec: dict):
+        for k in dataspec:
+            if '_dir' in k:
+                path = _os.path.join(self.args['dataset_dir'], dataspec[k])
+                path = path.replace(f"{_sep}{_sep}", _sep)
+                if path.endswith(_sep):
+                    path = path[:-1]
+                dataspec[k] = path
 
 
 class ETDataset(_Dataset):
     def __init__(self, mode='init', limit=None, **kw):
         self.mode = mode
         self.limit = limit
-        self.dataspecs = {}
         self.indices = []
+        self.data = {}
+
+        self.args = _etutils.FrozenDict(kw)
+        self.dataspecs = _etutils.FrozenDict({})
 
     def load_index(self, dataset_name, file):
         r"""
@@ -150,18 +211,50 @@ class ETDataset(_Dataset):
         """
         self.indices.append([dataset_name, file])
 
-    def _load_indices(self, dataset_name, files, **kw):
+    def _load_indices(self, dataspec_name, files, verbose=True):
         r"""
         We load the proper indices/names(whatever is called) of the files in order to prepare minibatches.
         Only load lim numbr of files so that it is easer to debug(Default is infinite, -lim/--load-lim argument).
         """
-        for file in files:
-            if self.limit and len(self) >= self.limit:
-                break
-            self.load_index(dataset_name, file)
+        _files = []
+        for i, f in enumerate(files[:self.args['load_limit']], 1):
+            _files.append([i, f])
 
-        if kw.get('verbose', True):
-            success(f'{dataset_name}, {self.mode}, {len(self)} Indices Loaded')
+        if len(_files) > 1:
+            nw = min(num_workers(self.args, self.args, self.args['use_ddp']), len(_files))
+            with _mp.Pool(processes=max(1, nw)) as pool:
+                all_datasets = list(
+                    _chain.from_iterable(pool.starmap(
+                        _partial(
+                            load_sparse_data_worker, self.__class__, self.mode, self.args,
+                            self.dataspecs[dataspec_name], len(_files)
+                        ),
+                        _files
+                    )))
+                self.gather_datasets_(all_datasets)
+
+
+        else:
+            for _, file in _files:
+                self.load_index(dataspec_name, file)
+
+        success(f'\n{dataspec_name}, {self.mode}, {len(self)} Indices Loaded', verbose and len(_files) > 1)
+
+    def gather_datasets_(self, dataset_objs):
+        for d in dataset_objs:
+            attributes = vars(d)
+            for k, v in attributes.items():
+                if isinstance(v, _etutils.FrozenDict):
+                    continue
+
+                if isinstance(v, list):
+                    self.__getattribute__(f"{k}").extend(v)
+
+                elif isinstance(attributes[f"{k}"], dict):
+                    self.__getattribute__(f"{k}").update(**v)
+
+                elif isinstance(attributes[f"{k}"], set):
+                    self.__getattribute__(f"{k}").union(v)
 
     def __getitem__(self, index):
         r"""
@@ -176,10 +269,10 @@ class ETDataset(_Dataset):
     def transforms(self, **kw):
         return None
 
-    def add(self, files, **kw):
+    def add(self, files, verbose=True, **kw):
         r""" An extra layer for added flexibility."""
         self.dataspecs[kw['name']] = kw
-        self._load_indices(dataset_name=kw['name'], files=files, verbose=kw.get('verbose'))
+        self._load_indices(dataspec_name=kw['name'], files=files, verbose=verbose)
 
     @classmethod
     def pool(cls, args, dataspecs, split_key=None, load_sparse=False):
@@ -192,15 +285,21 @@ class ETDataset(_Dataset):
         for dspec in dataspecs:
             for split in sorted(_os.listdir(dspec['split_dir'])):
                 split = _json.loads(open(dspec['split_dir'] + _os.sep + split).read())
+
+                _files = []
+                for i, f in enumerate(split[split_key][:args['load_limit']], 1):
+                    _files.append([i, f])
+
                 if load_sparse:
-                    for file in split[split_key]:
-                        if args['load_limit'] and len(all_d) >= args['load_limit']:
-                            break
-                        d = cls(mode=split_key, **args)
-                        d.add(files=[file], verbose=False, **dspec)
-                        all_d.append(d)
-                    if args['verbose']:
-                        success(f'{len(all_d)} sparse dataset loaded.')
+                    nw = min(num_workers(args, args, args['use_ddp']), len(_files))
+                    with _mp.Pool(processes=max(1, nw)) as pool:
+                        all_d += list(
+                            _chain.from_iterable(pool.starmap(
+                                _partial(
+                                    load_sparse_data_worker, cls, split_key, args, dspec, len(_files)
+                                ),
+                                _files
+                            )))
                 else:
                     if len(all_d) <= 0:
                         all_d.append(cls(mode=split_key, limit=args['load_limit'], **args))
@@ -208,6 +307,7 @@ class ETDataset(_Dataset):
                 """Pooling only works with 1 split at the moment."""
                 break
 
+        success(f'\nPooled {len(all_d)} dataset loaded.', args['verbose'] and len(all_d) > 1)
         return all_d
 
 
